"""
Complete implementation of the XDG Desktop Entry Specification Version 0.9.4
http://www.freedesktop.org/standards/desktop-entry-spec/

Not supported:
- Encoding: Legacy Mixed
- Does not check exec parameters
- Does not check if regex is valid
- Does not check if URL is valid
- Does not completly validate deprecated/kde items
- Does not check Category entries
- Does not check OnlyShowIn entries

TODO:
- Check for double Keys / Groups
- Get Section without []
"""

import string, codecs, re
from Exceptions import *

class DesktopEntry:
	"Class to parse and validate DesktopEntries"

	__content = {}
	__defaultGroup = ''
	__fileExtension = ''

	__type = ''
	__name = ''
	__version = ''
	__encoding = ''

	__locale_LANG = 'C'
	__locale_COUNTRY = ''
	__locale_ENCODING = ''
	__locale_MODIFIER = ''

	__file = ''

	__errors = []
	__warnings = []

	def __init__(self):
		# reset content
		self.__content = dict()
	
	def __str__(self):
		return self.getName()
	
	def __cmp__(self, other):
		return cmp(self.getName(), other.getName())

	def parse(self, file):
		# check file extension
		if re.search('\.desktop$', file):
			self.__fileExtension = 'desktop'
		elif re.search('\.directory$', file):
			self.__fileExtension = 'directory'

		# parse file
		try:
			f = codecs.open(file, 'r', 'utf-8')
			self.__file = file
		except IOError:
			raise ParsingError('File not found', file)

		try:
			lines = f.readlines()
		except UnicodeError:
			raise ParsingError('File contains non UTF-8 chars', file)

		for line in lines:
			# empty line
			if line.strip() == '':
				pass
			# comment
			elif line[0] == '#':
				pass
			# new group
			elif line[0] == '[':
				currentGroup = string.strip(line)
				self.__content[currentGroup] = {}
			# key
			else:
				tmp = string.split(line, '=')
				tmp[0] = string.strip(tmp[0])
				tmp[1] = string.strip(tmp[1])
				self.__content[currentGroup][tmp[0]] = tmp[1]

		# check header
		if self.__content.has_key('[KDE Desktop Entry]'):
			self.__defaultGroup = '[KDE Desktop Entry]'
		elif self.__content.has_key('[Desktop Entry]'):
			self.__defaultGroup = '[Desktop Entry]'
		else:
			raise ParsingError('[Desktop Entry]-Header missing', file)
	
	def getKey(self, key, group = '', locale = True):
		# set default group
		if not group:
			group = self.__defaultGroup

		# does Group exists?
		if not self.__content.has_key(group):
			raise NoGroupError(group, self.__file)

		# does key exists?
		if not self.__content[group].has_key(key):
			raise NoKeyError(key, group, self.__file)

		# return key (with locale)
		if locale:
			return self.__content[group][self.__addLocale(key, group)]
		else:
			return self.__content[group][key]

	def hasGroup(self, group):
		if self.__content.has_key(group):
			return True
		else:
			return False
	
	def hasKey(self, key, group = ''):
		if not group:
			group = self.__defaultGroup

		if self.hasGroup(group):
			if self.__content[group].has_key(key):
				return True
			else:
				return False
		else:
			return False

	def getActionKey(self, key, action, locale = False):
		return self.getKey(key, '[Desktop Action %s]' % action, locale)
	
	def getFileName(self):
		return self.__file

	# start standard keys
	def getType(self):
		return self.getKey('Type', False)
	def getVersion(self):
		return self.getKey('Version', False)
	def getEncoding(self):
		return self.getKey('Encoding', False)
	def getName(self):
		return self.getKey('Name')
	def getGenericName(self):
		return self.getKey('GenericName')
	def getComment(self):
		return self.getKey('Comment')
	def getNoDisplay(self):
		return self.__getBoolean(self.getKey('NoDisplay', False))
	def getIcon(self):
		# XXX: maybe implement the icon spec here ?
		return self.getKey('Icon')
	def getHidden(self):
		return self.__getBoolean(self.getKey('Hidden', False))
	def getFilePattern(self):
		return self.__getList(self.getKey('FilePattern', False))
	def getTryExec(self):
		return self.getKey('TryExec', False)
	def getExec(self):
		return self.getKey('Exec', False)
	def getPath(self):
		return self.getKey('Path', False)
	def getTerminal(self):
		return self.__getBoolean(self.getKey('Terminal', False))
	def getSwallowTitle(self):
		return self.getKey('SwallowTitle')
	def getSwallowExec(self):
		return self.getKey('SwallowExec', False)
	def getActions(self):
		return self.__getList(self.getKey('Actions', False))
	def getMimeType(self):
		return self.__getList(self.getKey('MimeType', False))
	def getSortOrder(self):	
		return self.__getList(self.getKey('SortOrder', False))
	def getDev(self):
		return self.getKey('Dev', False)
	def getFSType(self):
		return self.getKey('FSType', False)
	def getMountPoint(self):
		return self.getKey('MountPoint', False)
	def getReadonly(self):
		return self.__getBoolean(self.getKey('ReadOnly', False))
	def getUnmountIcon(self):
		# XXX: maybe implement the icon spec here ?
		return self.getKey('UnmountIcon')
	def getURL(self):
		return self.getKey('URL', False)
	def getCategories(self):
		return self.__getList(self.getKey('Categories', False))
	def getOnlyShowIn(self):
		return self.__getList(self.getKey('OnlyShowIn', False))
	def getNotShowIn(self):
		return self.__getList(self.getKey('NotShowIn', False))
	def getStartupNotify(self):
		return self.__getBoolean(self.getKey('StartupNotify', False))
	def getStartupWMClass(self):
		return self.getKey('StartupWMClass', False)
	# end standard keys

	# start kde keys
	def getServiceTypes(self):
		return self.__getList(self.getKey('ServiceTypes', False))
	def getDocPath(self):
		return self.getKey('DocPath', False)
	def getKeywords(self):
		return self.__getList(self.getKey('Keywords'))
	def getInitialPreference(self):
		return self.getKey('InitialPreference', False)
	# end kde keys

	# start deprecated keys
	def getMiniIcon(self):
		# XXX: maybe implement the icon spec here ?
		return self.getKey('MiniIcon')
	def getTerminalOptions(self):
		return self.getKey('TerminalOptions', False)
	def getDefaultApp(self):
		return self.getKey('DefaultApp', False)
	def getProtocols(self):
		return self.__getList(self.getKey('Protocols', False))
	def getExtensions(self):
		return self.__getList(self.getKey('Extensions', False))
	def getBinaryPattern(self):
		return self.getKey('BinaryPattern', False)
	def getMapNotify(self):
		return self.getKey('MapNotify', False)
	# end deprecated keys

	# start subget
	def __getList(self, string):
		if re.search(',', string):
			return string.split(',')
		else:
			return string.split(';')

	def __getBoolean(self, boolean):
		if boolean == 1:
			return True
		elif boolean == 0:
			return False
		else:
			return string.capitalize(boolean)
	# end subget

	# locale [] options
	def setLocale(self, lc_messages):
		"set locale for current desktop entry"
		# valid lc_messages format?
		p = re.compile('^([a-zA-Z]+)(_[a-zA-Z]+)?(\.[a-zA-Z\-0-9]+)?(@[a-zA-Z]+)?$')
		m = p.match(lc_messages)
		if not m:
			raise ValueError('Invalid LC_MESSAGES value')
		# set lc_messages
		else:
			self.__locale_LANG     = m.group(1)
			self.__locale_COUNTRY  = string.replace(m.group(2) or '', '_', '')
			self.__locale_ENCODING = string.replace(m.group(3) or '', '.', '')
			self.__locale_MODIFIER = string.replace(m.group(4) or '', '@', '')
	
	def getLocale(self):
		"return locale in lc_messages format"
		lc_messages = self.__locale_LANG
		if self.__locale_COUNTRY:
			lc_messages = lc_messages + '_' + self.__locale_COUNTRY
		if self.__locale_ENCODING:
			lc_messages = lc_messages + '.' + self.__locale_ENCODING
		if self.__locale_MODIFIER:
			lc_messages = lc_messages + '@' + self.__locale_MODIFIER
		return lc_messages

	def __addLocale(self, key, group = ''):
		"add locale to key according the current lc_messages"
		# set default group
		if not group:
			group = self.__defaultGroup

		# add locale
		if (self.__locale_LANG and self.__locale_COUNTRY and self.__locale_MODIFIER) and \
		(self.__content[group].has_key(key+'['+self.__locale_LANG+'_'+self.__locale_COUNTRY+'@'+self.__locale_MODIFIER+']')):
			return key+'['+self.__locale_LANG+'_'+self.__locale_COUNTRY+'@'+self.__locale_MODIFIER+']'
		elif (self.__locale_LANG and self.__locale_COUNTRY) and \
		(self.__content[group].has_key(key+'['+self.__locale_LANG+'_'+self.__locale_COUNTRY+']')):
			return key+'['+self.__locale_LANG+'_'+self.__locale_COUNTRY+']'
		elif (self.__locale_LANG and self.__locale_MODIFIER) and \
		(self.__content[group].has_key(key+'['+self.__locale_LANG+'@'+self.__locale_MODIFIER+']')):
			return key+'['+self.__locale_LANG+'@'+self.__locale_MODIFIER+']'
		elif (self.__locale_LANG) and \
		(self.__content[group].has_key(key+'['+self.__locale_LANG+']')):
			return key+'['+self.__locale_LANG+']'
		else:
			return key
	
	# validate ... (report = All / Warnings / Errrors)
	def validate(self, report = "All"):
		self.__errors = []
		self.__warnings = []

		# header
		if self.__defaultGroup == "[KDE Desktop Entry]":
			self.__warnings.append('[KDE Desktop Entry]-Header is deprecated')

		# file extension
		if not self.__fileExtension:
			self.__warnings.append('Unknown File extension')

		# check required keys
		# Type
		try:
			self.__type = self.__content[self.__defaultGroup]["Type"]
		except KeyError:
			self.__errors.append("Key 'Type' is missing")

		# Encoding
		try:
			self.__encoding = self.__content[self.__defaultGroup]["Encoding"]
		except KeyError:
			self.__errors.append("Key 'Encoding' is missing")

		# Version
		try:
			self.__version = self.__content[self.__defaultGroup]["Version"]
		except KeyError:
			self.__warnings.append("Key 'Version' is missing")

		# Name
		try:
			self.__name = self.__content[self.__defaultGroup]["Name"]
		except KeyError:
			self.__errors.append("Key 'Name' is missing")
		
		# check remaining keys
		for group in self.__content:
			self.__checkGroup(group)
			for key in self.__content[group]:
				self.__checkKey(key, self.__content[group][key])

		# raise Warnings / Errors
		msg = ""
		if report == "All" or report == "Warnings":
			for line in self.__warnings:
				msg += "\n- " + line

		if report == "All" or report == "Errors":
			for line in self.__errors:
				msg += "\n- " + line

		if msg:
			raise ValidationError(msg, self.__file)

	def __checkGroup(self, group):
		# check if group header is valid
		if not (group == self.__defaultGroup \
		or re.match("^\[Desktop Action [a-zA-Z]+\]$", group) \
		or (re.match("^\[X-", group) and group.encode("ascii", 'ignore') == group)):
			self.__errors.append("Invalid Group name: %s" % group.encode("ascii", "replace"))
		else:
		    #OnlyShowIn and NotShowIn
		    if self.__content[group].has_key("OnlyShowIn") and self.__content[group].has_key("NotShowIn"):
			self.__errors.append("Group may either have OnlyShowIn or NotShowIn, but not both")

	def __checkKey(self, key, value):
		locale = "(\[([a-zA-Z]+)(_[a-zA-Z]+)?(\.[a-zA-Z\-0-9]+)?(@[a-zA-Z]+)?\])?"

		# standard keys		
		if key == "Type":
			if value == "ServiceType" or value == "Service":
				self.__warnings.append("Type=%s is a KDE extension" % key)
			elif value == "MimeType":
				self.__warnings.append("Type=MimeType is deprecated")
			elif not (value == "Application" or value == "Link" or value == "FSDevice" or value == "Directory"):
				self.__errors.append("Value of key 'Type' must be Application, Link, FSDevice or Directory, but is '%s'" % value)

			if self.__fileExtension == "directory" and not value == "Directory":
				self.__warnings.append("File extension is .directory, but Type is '%s'" % value)
			elif self.__fileExtension == "desktop" and value == "Directory":
				self.__warnings.append("Files with Type=Directory should have the extension .directory")

		elif key == "Version":
			self.__checkNumber(key, value)

		elif key == "Encoding":
			if value == "Legacy-Mixed":
				self.__errors.append("Encoding=Legacy-Mixed is deprecated and not supported by this parser")
			elif not value == "UTF-8":
				self.__errors.append("Value of key 'Encoding' must be UTF-8")

		elif re.match("^Name"+locale+"$", key):
			pass # locale string

		elif re.match("^GenericName"+locale+"$", key):
			pass # locale string

		elif re.match("^Comment"+locale+"$", key):
			pass # locale string

		elif key == "NoDisplay":
			self.__checkBoolean(key, value)

		elif key == "Hidden":
			self.__checkBoolean(key, value)

		elif key == "Terminal":
			self.__checkBoolean(key, value)
			self.__checkType(key, "Application")

		elif key == "TryExec":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif key == "Exec":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif key == "Path":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif re.match("^Icon"+locale+"$", key):
			# XXX: maybe implement the icon spec here ?
			self.__checkString(key, value)
		
		elif re.match("^SwallowTitle"+locale+"$", key):
			self.__checkType(key, "Application")

		elif key == "SwallowExec":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif key == "FilePatterns":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif key == "Actions":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif key == "MimeType":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif key == "Categories":
			self.__checkString(key, value)
			self.__checkType(key, "Application")

		elif key == "OnlyShowIn":
			self.__checkString(key, value)

		elif key == "NotShowIn":
			self.__checkString(key, value)

		elif key == "StartupNotify":
			self.__checkBoolean(key, value)
			self.__checkType(key, "Application")

		elif key == "StartupWMClass":
			self.__checkType(key, "Application")

		elif key == "SortOrder":
			self.__checkString(key, value)
			self.__checkType(key, "Directory")

		elif key == "URL":
			self.__checkString(key, value)
			self.__checkType(key, "URL")

		elif key == "Dev":
			self.__checkString(key, value)
			self.__checkType(key, "FSDevice")

		elif key == "FSType":
			self.__checkString(key, value)
			self.__checkType(key, "FSDevice")

		elif key == "MountPoint":
			self.__checkString(key, value)
			self.__checkType(key, "FSDevice")

		elif re.match("^UnmountIcon"+locale+"$", key):
			# XXX: maybe implement the icon spec here ?
			self.__checkString(key, value)
			self.__checkType(key, "FSDevice")

		elif key == "ReadOnly":
			self.__checkBoolean(key, value)
			self.__checkType(key, "FSDevice")

		# kde extensions
		elif key == "ServiceTypes":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is a KDE extension" % key)

		elif key == "DocPath":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is a KDE extension" % key)

		elif re.match("^Keywords"+locale+"$", key):
			self.__warnings.append("Key '%s' is a KDE extension" % key)

		elif key == "InitialPreference":
			self.__checkNumber(key, value)
			self.__warnings.append("Key '%s' is a KDE extension" % key)

		# deprecated keys
		elif re.match("^MiniIcon"+locale+"$", key):
			# XXX: maybe implement the icon spec here ?
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is deprecated" % key)

		elif key == "TerminalOptions":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is deprecated" % key)

		elif key == "DefaultApp":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is deprecated" % key)

		elif key == "Protocols":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is deprecated" % key)

		elif key == "^Extensions":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is deprecated" % key)

		elif key == "BinaryPattern":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is deprecated" % key)

		elif key == "MapNotify":
			self.__checkString(key, value)
			self.__warnings.append("Key '%s' is deprecated" % key)

		# "X-" extensions
		elif re.match("^X-[a-zA-Z0-9-]+", key):
			pass

		else:
			self.__errors.append("Invalid key: %s" % key)

		# check if value is empty
		if value == "":
			self.__warnings.append("Value of Key '%s' is empty" % key)

	def __checkBoolean(self, key, value):
		# 1 or 0 : deprecated
		if (value == "1" or value == "0"):
			self.__warnings.append("Value of key '%s' is a deprecated boolean value" % key)
		# true or false: ok
		elif not (value == "true" or value == "false"):
			self.__errors.append("Value of key '%s' is not a boolean" % key)

	def __checkNumber(self, key, value):
		# float() ValueError
		try:
			float(value)
		except:
			self.__errors.append("Value of key '%s' is not a number" % key)

	def __checkString(self, key, value):
		# convert to ascii
		if not value.encode("ascii", 'ignore') == value:
			self.__errors.append("Value of key '%s' must only contain ASCII characters" % key)

	def __checkType(self, key, type):
		if not self.__type == type:
			self.__errors.append("Key '%s' only allowed in Type=%s" % (key, type))
